/* Copyright 2013-present Barefoot Networks, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * Antonin Bas (antonin@barefootnetworks.com)
 *
 */

#include <bm/bm_sim/_assert.h>
#include <bm/bm_sim/calculations.h>
#include <bm/bm_sim/logger.h>
#include <bm/bm_sim/packet.h>
#include <bm/bm_sim/phv.h>

#include <netinet/in.h>

#include <algorithm>
#include <mutex>
#include <ostream>
#include <string>
#include <unordered_map>

#include "xxhash.h"
#include "crc_tables.h"
#include "extract.h"

namespace bm {

void
BufBuilder::push_back_field(header_id_t header, int field_offset) {
  field_t f = {header, field_offset};
  entries.emplace_back(f);
}

void
BufBuilder::push_back_constant(const ByteContainer &v, size_t nbits) {
  constant_t c = {v, nbits};
  entries.emplace_back(c);
}

void
BufBuilder::push_back_header(header_id_t header) {
  header_t h = {header};
  entries.emplace_back(h);
}

void
BufBuilder::append_payload() {
  with_payload = true;
}

struct BufBuilder::Deparse : public boost::static_visitor<> {
  Deparse(const PHV &phv, ByteContainer *buf, int init_offset = 0)
      : phv(phv), buf(buf), nbits(init_offset) { }

  char *extend(int more_bits) {
    int nbits_ = nbits + more_bits;
    buf->resize((nbits_ + 7) / 8, '\x00');
    char *ptr = buf->data() + (nbits / 8);
    nbits = nbits_;
    return ptr;
  }

  int get_offset() const {
    return nbits % 8;
  }

  void operator()(const field_t &f) {
    const Header &header = phv.get_header(f.header);
    if (!header.is_valid()) return;
    const Field &field = header.get_field(f.field_offset);
    // get_offset needs to be called before extend!
    const auto offset = get_offset();
    field.deparse(extend(field.get_nbits()), offset);
  }

  void operator()(const constant_t &c) {
    // get_offset needs to be called before extend!
    const auto offset = get_offset();
    extract::generic_deparse(&c.v[0], c.nbits, extend(c.nbits), offset);
  }

  void operator()(const header_t &h) {
    assert(get_offset() == 0);
    const Header &header = phv.get_header(h.header);
    if (header.is_valid()) {
      header.deparse(extend(header.get_nbytes_packet() * 8));
    }
  }

  Deparse(const Deparse &other) = delete;
  Deparse &operator=(const Deparse &other) = delete;

  Deparse(Deparse &&other) = delete;
  Deparse &operator=(Deparse &&other) = delete;

  const PHV &phv;
  ByteContainer *buf;
  int nbits{0};
};

void
BufBuilder::operator()(const Packet &pkt, ByteContainer *buf) const {
  buf->clear();
  const PHV *phv = pkt.get_phv();
  Deparse visitor(*phv, buf);
  std::for_each(entries.begin(), entries.end(),
                boost::apply_visitor(visitor));
  if (with_payload) {
    size_t curr = buf->size();
    size_t psize = pkt.get_data_size();
    buf->resize(curr + psize);
    std::copy(pkt.data(), pkt.data() + psize, buf->begin() + curr);
  }
}

namespace hash {

uint64_t xxh64(const char *buffer, size_t s) {
  return XXH64(buffer, s, 0);
}

}  // namespace hash

namespace {

/* This code was adapted from:
   http://www.barrgroup.com/Embedded-Systems/How-To/CRC-Calculation-C-Code */

template <typename T>
T reflect(T data, int nBits) {
  T reflection = static_cast<T>(0x00);
  int bit;

  // Reflect the data about the center bit.
  for (bit = 0; bit < nBits; ++bit) {
    // If the LSB bit is set, set the reflection of it.
    if (data & 0x01) {
      reflection |= (static_cast<T>(1) << ((nBits - 1) - bit));
    }
    data = (data >> 1);
  }

  return reflection;
}

struct xxh64 {
  uint64_t operator()(const char *buf, size_t len) const {
    return XXH64(buf, len, 0);
  }
};

struct crc16 {
  uint16_t operator()(const char *buf, size_t len) const {
    uint16_t remainder = 0x0000;
    uint16_t final_xor_value = 0x0000;
    for (unsigned int byte = 0; byte < len; byte++) {
      int data = reflect<uint16_t>(buf[byte], 8) ^ (remainder >> 8);
      remainder = table_crc16[data] ^ (remainder << 8);
    }
    return reflect<uint16_t>(remainder, 16) ^ final_xor_value;
  }
};

template <typename T>
struct crc_custom_init { };

// explicit template instantiation

// crc8

template <>
struct crc_custom_init<uint8_t> {
  static CustomCrcMgr<uint8_t>::crc_config_t config;
};

CustomCrcMgr<uint8_t>::crc_config_t crc_custom_init<uint8_t>::config =
{0x07, 0x00, 0x00, false, false};

// crc16

template <>
struct crc_custom_init<uint16_t> {
  static CustomCrcMgr<uint16_t>::crc_config_t config;
};

CustomCrcMgr<uint16_t>::crc_config_t crc_custom_init<uint16_t>::config =
{0x8005, 0x0000, 0x0000, true, true};

// crc32

template <>
struct crc_custom_init<uint32_t> {
  static CustomCrcMgr<uint32_t>::crc_config_t config;
};

CustomCrcMgr<uint32_t>::crc_config_t crc_custom_init<uint32_t>::config =
{0x04c11db7, 0xffffffff, 0xffffffff, true, true};

// crc64

template <>
struct crc_custom_init<uint64_t> {
  static CustomCrcMgr<uint64_t>::crc_config_t config;
};

CustomCrcMgr<uint64_t>::crc_config_t crc_custom_init<uint64_t>::config =
{0x000000000000001bULL, 0x0000000000000000ULL, 0x0000000000000000ULL,
true, true};

template <typename T>
struct crc_custom {
  static constexpr size_t width = sizeof(T) * 8;
  static constexpr size_t kTEntries = 256u;
  using crc_config_t = typename CustomCrcMgr<T>::crc_config_t;

  crc_custom() {
    config = crc_custom_init<T>::config;
    recompute_crc_table(config, crc_table);
  }

  T operator()(const char *buf, size_t len) const {
    // clearly not optimized (critical section may be made smaller), but we will
    // try to do better if needed
    std::unique_lock<std::mutex> lock(m);

    T remainder = config.initial_remainder;
    for (unsigned int byte = 0; byte < len; byte++) {
      unsigned char uchar = static_cast<unsigned char>(buf[byte]);
      int data = (config.data_reflected) ?
          reflect<T>(uchar, 8) ^ (remainder >> (width - 8)) :
          uchar ^ (remainder >> (width - 8));
      remainder = crc_table[data] ^ (remainder << 8);
    }
    return (config.remainder_reflected) ?
        reflect<T>(remainder, width) ^ config.final_xor_value :
        remainder ^ config.final_xor_value;
  }

  void update_config(const crc_config_t &new_config) {
    T crc_table_new[kTEntries];
    recompute_crc_table(new_config, crc_table_new);

    std::unique_lock<std::mutex> lock(m);
    config = new_config;
    std::memcpy(crc_table, crc_table_new, sizeof(crc_table));
  }

 private:
  void recompute_crc_table(const crc_config_t &new_config, T *new_table) {
    // Compute the remainder of each possible dividend
    for (size_t dividend = 0; dividend < kTEntries; dividend++) {
      // Start with the dividend followed by zeros
      T remainder = static_cast<T>(dividend) << (width - 8);
      // Perform modulo-2 division, a bit at a time
      for (unsigned char bit = 8; bit > 0; bit--) {
        // Try to divide the current data bit
        if (remainder & (static_cast<T>(1) << (width - 1))) {
          remainder = (remainder << 1) ^ new_config.polynomial;
        } else {
          remainder = (remainder << 1);
        }
      }

      // Compute the remainder of each possible dividend
      new_table[dividend] = remainder;
    }
  }

  T crc_table[kTEntries];
  crc_config_t config;
  mutable std::mutex m{};
};

struct crc32 {
  uint32_t operator()(const char *buf, size_t len) const {
    uint32_t remainder = 0xFFFFFFFF;
    uint32_t final_xor_value = 0xFFFFFFFF;
    for (unsigned int byte = 0; byte < len; byte++) {
      int data = reflect<uint32_t>(buf[byte], 8) ^ (remainder >> 24);
      remainder = table_crc32[data] ^ (remainder << 8);
    }
    return reflect<uint32_t>(remainder, 32) ^ final_xor_value;
  }
};

struct crcCCITT {
  uint16_t operator()(const char *buf, size_t len) const {
    uint16_t remainder = 0xFFFF;
    uint16_t final_xor_value = 0x0000;
    for (unsigned int byte = 0; byte < len; byte++) {
      int data = static_cast<unsigned char>(buf[byte]) ^ (remainder >> 8);
      remainder = table_crcCCITT[data] ^ (remainder << 8);
    }
    return remainder ^ final_xor_value;
  }
};

struct cksum16 {
  uint16_t operator()(const char *buf, size_t len) const {
    uint64_t sum = 0;
    const uint64_t *b = reinterpret_cast<const uint64_t *>(buf);
    uint32_t t1, t2;
    uint16_t t3, t4;
    const uint8_t *tail;
    /* Main loop - 8 bytes at a time */
    while (len >= sizeof(uint64_t)) {
      uint64_t s = *b++;
      sum += s;
      if (sum < s) sum++;
      len -= 8;
    }
    /* Handle tail less than 8-bytes long */
    tail = reinterpret_cast<const uint8_t *>(b);
    if (len & 4) {
      uint32_t s = *reinterpret_cast<const uint32_t *>(tail);
      sum += s;
      if (sum < s) sum++;
      tail += 4;
    }
    if (len & 2) {
      uint16_t s = *reinterpret_cast<const uint16_t *>(tail);
      sum += s;
      if (sum < s) sum++;
      tail += 2;
    }
    if (len & 1) {
      uint8_t s = *reinterpret_cast<const uint8_t *>(tail);
      sum += s;
      if (sum < s) sum++;
    }
    /* Fold down to 16 bits */
    t1 = sum;
    t2 = sum >> 32;
    t1 += t2;
    if (t1 < t2) t1++;
    t3 = t1;
    t4 = t1 >> 16;
    t3 += t4;
    if (t3 < t4) t3++;
    return ntohs(~t3);
  }
};

struct csum16 {
  uint16_t operator()(const char *buf, size_t len) const {
    return cksum16()(buf, len);
  }
};

struct xor16 {
  uint16_t operator()(const char *buf, size_t len) const {
    uint16_t mask = 0x00ff;
    uint16_t final_xor_value = 0x0000;
    unsigned int byte = 0;
    uint16_t t1, t2;
    /* Main loop - 2 bytes at a time */
    while (len >= 2) {
      t1 = static_cast<uint16_t>(buf[byte]) << 8;
      t2 = static_cast<uint16_t>(buf[byte + 1]);
      final_xor_value = final_xor_value ^ (t1 + (t2 & mask));
      byte += 2;
      len -= 2;
    }
    if (len > 0) {
      t1 = static_cast<uint16_t>(buf[byte]) << 8;
      final_xor_value = final_xor_value ^ t1;
    }
    return final_xor_value;
  }
};

struct identity {
  uint64_t operator()(const char *buf, size_t len) const {
    uint64_t res = 0ULL;
    for (size_t i = 0; i < std::min(sizeof(res), len); i++) {
      if (i > 0) res <<= 8;
      res += static_cast<uint8_t>(buf[i]);
    }
    return res;
  }
};

// This "hash" functor keeps an internal counter which is incremented by 1 every
// time it is called. It does not do any hashing but can be convenient to
// iterate through all possible scenari when a modulo is applied to the result
// of the hash (e.g. for action profile member selection).
// Different instances of the hash use different counters. As a remainder each
// instance can be executed from multiple threads (e.g. for traffic on different
// ports), thus the mutex. All the state is declared as "mutable" since we
// require that the functor be "const".
struct round_robin {
  uint64_t operator()(const char *buf, size_t len) const {
    _BM_UNUSED(buf);
    _BM_UNUSED(len);
    std::unique_lock<std::mutex> lock(m);
    return idx++;
  }

  mutable uint64_t idx{0};
  mutable std::mutex m{};
};

// Same as "round_robin" above but keeps an internal map to ensure that multiple
// calls to the functor with the same input yield the same result.
struct round_robin_consistent {
  uint64_t operator()(const char *buf, size_t len) const {
    std::unique_lock<std::mutex> lock(m);
    std::string hash(buf, len);
    auto p = hashes.emplace(hash, idx);
    // p.second == true <=> insertion took place
    return (p.second) ? (idx++) : p.first->second;
  }

  mutable uint64_t idx{0};
  mutable std::unordered_map<std::string, uint64_t> hashes;
  mutable std::mutex m{};
};

}  // namespace

// if REGISTER_HASH calls placed in the anonymous namespace, some compiler can
// give an unused variable warning
REGISTER_HASH(xxh64);
REGISTER_HASH(crc16);
REGISTER_HASH(crc32);
REGISTER_HASH(crcCCITT);
REGISTER_HASH(cksum16);
REGISTER_HASH(csum16);
REGISTER_HASH(identity);
REGISTER_HASH(round_robin);
REGISTER_HASH(round_robin_consistent);
REGISTER_HASH(xor16);

using crc8_custom = crc_custom<uint8_t>;
REGISTER_HASH(crc8_custom);
using crc16_custom = crc_custom<uint16_t>;
REGISTER_HASH(crc16_custom);
using crc32_custom = crc_custom<uint32_t>;
REGISTER_HASH(crc32_custom);
using crc64_custom = crc_custom<uint64_t>;
REGISTER_HASH(crc64_custom);

namespace detail {

template <typename T>
std::ostream &operator<<(std::ostream &out, const crc_config_t<T> &c) {
  out << "polynomial: " << c.polynomial << ", initial_remainder: "
      << c.initial_remainder << ", final_xor_value: " << c.final_xor_value
      << ", data_reflected: " << c.data_reflected
      << ", remainder_reflected: " << c.remainder_reflected;
  return out;
}

}  // namespace detail

template <typename T>
CustomCrcErrorCode
CustomCrcMgr<T>::update_config(NamedCalculation *calculation,
                               const crc_config_t &config) {
  Logger::get()->info("Updating config of custom crc {}: {}",
                      calculation->get_name(), config);
  auto raw_c_iface = calculation->get_raw_calculation();
  return update_config(raw_c_iface, config);
}

template <typename T>
CustomCrcErrorCode
CustomCrcMgr<T>::update_config(RawCalculationIface<uint64_t> *c,
                               const crc_config_t &config) {
  using ExpectedCType = RawCalculation<uint64_t, crc_custom<T> >;
  auto raw_c = dynamic_cast<ExpectedCType *>(c);
  if (!raw_c) return CustomCrcErrorCode::WRONG_TYPE_CALCULATION;
  raw_c->get_hash_fn().update_config(config);
  return CustomCrcErrorCode::SUCCESS;
}

// explicit instantiation, should prevent implicit instantiation
template class CustomCrcMgr<uint8_t>;
template class CustomCrcMgr<uint16_t>;
template class CustomCrcMgr<uint32_t>;
template class CustomCrcMgr<uint64_t>;

CalculationsMap * CalculationsMap::get_instance() {
  static CalculationsMap map;
  return &map;
}

bool CalculationsMap::register_one(const char *name, std::unique_ptr<MyC> c) {
  const std::string str_name = std::string(name);
  auto it = map_.find(str_name);
  if (it != map_.end()) return false;
  map_[str_name] = std::move(c);
  return true;
}

std::unique_ptr<CalculationsMap::MyC>
CalculationsMap::get_copy(const std::string &name) {
  auto it = map_.find(name);
  if (it == map_.end()) return nullptr;
  return it->second->clone();
}

}  // namespace bm
